import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.common.notable.enums import (
    NUMBERING_ORDER,
    Color,
    RunStatus,
    TaskStatus,
    LabwareType,
    Height,
    Plunger,
    LocationType,
    LocationReference,
)
from ablelabs.neon.utils.enum_types import Enum, auto, StrEnum, ItemGetable


class WasherSuctionValve(StrEnum):
    NEEDLE = auto()
    BATH = auto()


class Axis(StrEnum):
    X = auto()
    Y = auto()
    Z1 = auto()
    Z2 = auto()
    P1 = auto()
    P2 = auto()

    WDX = auto()
    WDY = auto()
    WDZ = auto()
    WDP = auto()

    MSZ = auto()
    MSS = auto()

    Z = ItemGetable[int](lambda number: Axis.from_str(f"{Axis.Z1[0]}{number}"))
    P = ItemGetable[int](lambda number: Axis.from_str(f"{Axis.P1[0]}{number}"))


class DioInput(StrEnum):
    DOOR = auto()


class DioPDO(StrEnum):
    LED_LAMP = auto()


class AxisDO(StrEnum):
    WASHER_DRYER_FAN = auto()
    WASHER_DRYER_VALVE = auto()
    WASHER_DRYER_SUCTION_PUMP = auto()



from typing import Callable, Coroutine, Literal

import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.network.messenger import MessengerServer
from ablelabs.neon.common.alma.enums import (
    DioInput,
    Axis,
    RunStatus,
    LocationType,
    ZGripper,
)
from ablelabs.neon.common.alma.structs import (
    Location,
    LCRParam,
    HeaterData,
    State,
    Speed,
)


class RobotRouter(MessengerServer):
    # robot
    async def robot_wait_boot(self):
        pass

    async def robot_stop(self):
        pass

    async def robot_clear_error(self):
        pass

    async def robot_pause(self):
        pass

    async def robot_resume(self):
        pass

    async def robot_is_connected(self) -> dict[str, bool]:
        pass

    async def robot_get_environment(self) -> dict[str, float]:
        pass

    async def robot_get_run_status(self) -> RunStatus:
        pass

    async def robot_get_dio_input(self) -> dict[DioInput, bool]:
        pass

    async def robot_get_last_state(self) -> State:
        pass

    async def robot_get_changed_robot_status(self) -> dict:
        pass

    async def robot_set_interlock(self, value: bool):
        pass

    async def robot_set_led_lamp(self, on: bool):
        pass

    async def robot_set_run_status(self, value: RunStatus):
        pass

    async def robot_set_progress_rate(self, value: float):
        pass

    async def robot_initialize(self, recovery: bool):
        pass

    async def robot_initialize_teaching(self):
        pass

    async def robot_calibrate_lcr_meter(self):
        pass

    # get
    async def get_setup_data(self) -> dict:
        pass

    # set
    async def set_update_setup_data(self, value: dict, save: bool):
        pass

    async def set_update_pipette_attrs(self, value: dict[int, dict]):
        pass

    # time
    async def time_initialize(self) -> float:
        pass

    async def time_calibrate_lcr_meter(self) -> float:
        pass

    # time.upper_module
    async def time_upper_module_move_z_up(self) -> float:
        pass

    async def time_upper_module_move_to_ready(self) -> float:
        pass

    # time.upper_module.gripper
    async def time_upper_module_gripper_move_block_to_inspector(
        self,
        push: bool,
    ) -> float:
        pass

    async def time_upper_module_gripper_move_block_to_holder(self) -> float:
        pass

    async def time_upper_module_gripper_open_lid(
        self,
        chip_holder_number: int,
        lid_holder_number: int,
    ) -> float:
        pass

    async def time_upper_module_gripper_close_lid(
        self,
        lid_holder_number: int,
        chip_holder_number: int,
    ) -> float:
        pass

    async def time_upper_module_gripper_move_labware(
        self,
        from_location: Location,
        to_location: Location,
        push: bool = False,
    ) -> float:
        pass

    async def time_upper_module_gripper_move_lid(
        self,
        from_location: Location,
        to_location: Location,
    ) -> float:
        pass

    async def time_upper_module_gripper_pick_labware(self, location: Location) -> float:
        pass

    async def time_upper_module_gripper_place_labware(
        self,
        location: Location,
        push: bool = False,
    ) -> float:
        pass

    async def time_upper_module_gripper_pick_lid(self, location: Location) -> float:
        pass

    async def time_upper_module_gripper_place_lid(self, location: Location) -> float:
        pass

    async def time_upper_module_gripper_move_to(
        self,
        location: Location,
        z_gripper: ZGripper,
        slow_z_offset: float = 0,
    ) -> float:
        pass

    # time.upper_module.pipette
    async def time_upper_module_pipette_pick_up_tip(
        self, well: str, optimize: bool = True
    ) -> float:
        pass

    async def time_upper_module_pipette_drop_tip(
        self,
        location_type: LocationType,
        well: str,
        optimize: bool = True,
    ) -> float:
        pass

    async def time_upper_module_pipette_move_to(
        self,
        location: Location,
        z_speed: Speed = None,
        optimize: bool = False,
    ) -> float:
        pass

    async def time_upper_module_pipette_ready_plunger(
        self, flow_rate: float = None
    ) -> float:
        pass

    async def time_upper_module_pipette_blow_out(
        self, flow_rate: float = None
    ) -> float:
        pass

    async def time_upper_module_pipette_aspirate(
        self,
        volume: float | list[float],  # list for multi-dispense calibration
        flow_rate: float = None,
        height_offset: float = 0,
    ) -> float:
        pass

    async def time_upper_module_pipette_dispense(
        self,
        volume: float,
        flow_rate: float = None,
        height_offset: float = 0,
    ) -> float:
        pass

    async def time_upper_module_pipette_mix(
        self,
        volume: float | list[float],  # list for multi-dispense calibration
        iteration: int,
        flow_rate: float = None,
        delay: float = 0.0,
        height_offset: float = 0,
    ) -> float:
        pass

    # upper_module
    async def upper_module_move_z_up(self):
        pass

    async def upper_module_move_to_ready(self):
        pass

    # upper_module.gripper
    async def upper_module_gripper_move_block_to_inspector(self, push: bool):
        pass

    async def upper_module_gripper_move_block_to_holder(self):
        pass

    async def upper_module_gripper_open_lid(
        self,
        chip_holder_number: int,
        lid_holder_number: int,
    ):
        pass

    async def upper_module_gripper_close_lid(
        self,
        lid_holder_number: int,
        chip_holder_number: int,
    ):
        pass

    async def upper_module_gripper_move_labware(
        self,
        from_location: Location,
        to_location: Location,
        push: bool = False,
    ):
        pass

    async def upper_module_gripper_move_lid(
        self,
        from_location: Location,
        to_location: Location,
    ):
        pass

    async def upper_module_gripper_pick_labware(self, location: Location):
        pass

    async def upper_module_gripper_place_labware(
        self,
        location: Location,
        push: bool = False,
    ):
        pass

    async def upper_module_gripper_pick_lid(self, location: Location):
        pass

    async def upper_module_gripper_place_lid(self, location: Location):
        pass

    async def upper_module_gripper_move_to(
        self,
        location: Location,
        z_gripper: ZGripper,
        task_after_z_up: Callable[[], Coroutine] = None,
        slow_z_offset: float = 0,
    ):
        pass

    # upper_module.pipette
    async def upper_module_pipette_pick_up_tip(self, well: str, optimize: bool = True):
        pass

    async def upper_module_pipette_drop_tip(
        self,
        location_type: LocationType,
        well: str,
        optimize: bool = True,
    ):
        pass

    async def upper_module_pipette_move_to(
        self,
        location: Location,
        task_after_z_up: Callable[[], Coroutine] = None,
        z_speed: Speed = None,
        optimize: bool = False,
    ):
        pass

    async def upper_module_pipette_ready_plunger(self, flow_rate: float = None):
        pass

    async def upper_module_pipette_blow_out(self, flow_rate: float = None):
        pass

    async def upper_module_pipette_aspirate(
        self,
        volume: float | list[float],  # list for multi-dispense calibration
        flow_rate: float = None,
        height_offset: float = 0,
    ):
        pass

    async def upper_module_pipette_dispense(
        self,
        volume: float,
        flow_rate: float = None,
        height_offset: float = 0,
    ):
        pass

    async def upper_module_pipette_mix(
        self,
        volume: float | list[float],  # list for multi-dispense calibration
        iteration: int,
        flow_rate: float = None,
        delay: float = 0.0,
        height_offset: float = 0,
    ):
        pass

    # deck_module.inspector
    async def deck_module_inspector_set_lcr_param(self, value: LCRParam):
        pass

    async def deck_module_inspector_scan(self) -> dict[int, float]:
        pass

    async def deck_module_inspector_prepare_scan(self):
        pass

    async def deck_module_inspector_complete_scan(self):
        pass

    async def deck_module_inspector_set_switch_on(self, channel: int):
        pass

    async def deck_module_inspector_set_switch_off(self):
        pass

    async def deck_module_inspector_get_lcr_value(self) -> float:
        pass

    # deck_module.heater
    async def deck_module_heater_get_heater_on(self) -> dict[int, bool]:
        pass

    async def deck_module_heater_get_temperature(self) -> dict[int, float]:
        pass

    async def deck_module_heater_set_environment_temperature(self, value: float):
        pass

    async def deck_module_heater_set_temperature(self, values: dict[int, HeaterData]):
        pass

    # axis
    async def axis_get_position(self, axis: Axis, floor_digit: int = 1) -> float:
        pass

    async def axis_set_speed(self, axis: Axis, value: float):
        pass

    async def axis_set_accel(self, axis: Axis, value: float):
        pass

    async def axis_set_decel(self, axis: Axis, value: float):
        pass

    async def axis_enable(self, axis: Axis):
        pass

    async def axis_disable(self, axis: Axis):
        pass

    async def axis_stop(self, axis: Axis):
        pass

    async def axis_home(self, axis: Axis):
        pass

    async def axis_jog(self, axis: Axis, value: float):
        pass

    async def axis_step(self, axis: Axis, value: float):
        pass

    async def axis_move(self, axis: Axis, value: float):
        pass

    async def axis_wait_home_done(self, axis: Axis):
        pass

    async def axis_wait_move_done(self, axis: Axis):
        pass



from typing import Any
import cv2

import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.network.messenger import MessengerServer
from ablelabs.neon.common.notable.enums import Axis
from ablelabs.neon.common.notable.structs import Location, Speed, FlowRate
from ablelabs.neon.common.notable.kribb.structs import LabelInfo


class RobotRouter(MessengerServer):
    async def robot_wait_boot(self):
        pass

    async def robot_stop(self):
        pass

    async def robot_clear_error(self):
        pass

    async def robot_pause(self):
        pass

    async def robot_resume(self):
        pass

    async def robot_is_connected(self):
        pass

    async def robot_scan_displacement_sensor(
        self,
        pipette_number: int,
        location: Location,
        padding: tuple[float, float],
        scan_count: tuple[float, float],
    ):
        pass

    async def robot_move_to_interpolated_z(
        self,
        pipette_number: int,
        location: Location,
        x: float,
        y: float,
    ):
        pass

    # set api
    async def set_pipettes(self, value: dict[int, str | None]):
        pass

    async def set_tips(self, value: dict[int, str | None]):
        pass

    async def set_deck_modules(self, value: dict[int, str | None]):
        pass

    async def set_labwares(self, value: dict[int, str | None]):
        pass

    async def set_update_pipette_attrs(self, value: dict[int, dict]):
        pass

    async def set_update_setup_data(self, value: dict, save: bool):
        pass

    async def set_camera(self, pipette_number: int):
        pass

    # get api
    async def get_pipette_infos(self):
        pass

    async def get_tip_infos(self):
        pass

    async def get_deck_module_infos(self):
        pass

    async def get_labware_infos(self):
        pass

    async def get_setup_data(self):
        pass

    # motion api
    async def motion_initialize(self):
        pass

    async def motion_delay(self, sec: float):
        pass

    async def motion_move_to_ready(self):
        pass

    async def motion_move_to(
        self,
        pipette_number: int,
        location: Location,
        optimize: bool = False,
    ):
        pass

    async def motion_move_to_camera(
        self,
        pipette_number: int,
        location: Location,
    ):
        pass

    async def motion_move_to_displacement_sensor(
        self,
        pipette_number: int,
        location: Location,
    ):
        pass

    async def motion_pick_up_tip(
        self,
        pipette_number: int,
        location: Location,
        optimize: bool = False,
    ):
        pass

    async def motion_drop_tip(
        self,
        pipette_number: int,
        location: Location,
        optimize: bool = False,
    ):
        pass

    async def motion_rise_tip(
        self,
        pipette_number: int,
        height_offset: float,
        z_speed: Speed,
    ):
        pass

    async def motion_aspirate(
        self,
        pipette_number: int,
        volume: float,
        location: Location = None,
        flow_rate: FlowRate = None,
        optimize: bool = False,
    ):
        pass

    async def motion_dispense(
        self,
        pipette_number: int,
        volume: float,
        location: Location = None,
        flow_rate: FlowRate = None,
        optimize: bool = False,
    ):
        pass

    async def motion_mix(
        self,
        pipette_number: int,
        volume: float,
        iteration: int,
        location: Location = None,
        flow_rate: FlowRate = None,
        delay: float = 0.0,
        optimize: bool = False,
    ):
        pass

    async def motion_blow_out(
        self,
        pipette_number: int,
        flow_rate: FlowRate = None,
    ):
        pass

    # axis api
    async def axis_get_position(self, axis: Axis, floor_digit: int = 1):
        pass

    async def axis_set_speed(self, axis: Axis, value: float):
        pass

    async def axis_set_accel(self, axis: Axis, value: float):
        pass

    async def axis_set_decel(self, axis: Axis, value: float):
        pass

    async def axis_enable(self, axis: Axis):
        pass

    async def axis_disable(self, axis: Axis):
        pass

    async def axis_stop(self, axis: Axis):
        pass

    async def axis_home(self, axis: Axis):
        pass

    async def axis_jog(self, axis: Axis, value: float):
        pass

    async def axis_step(self, axis: Axis, value: float):
        pass

    async def axis_move(self, axis: Axis, value: float):
        pass

    async def axis_wait_home_done(self, axis: Axis):
        pass

    async def axis_wait_move_done(self, axis: Axis):
        pass

    # optic api
    async def optic_set_camera_live(self, on: bool):
        pass

    async def optic_camera_capture(self, resize_ratio: float = None):
        pass

    async def optic_camera_show(
        self,
        winname: str,
        resize_ratio: float = 0.5,
    ):
        pass

    async def optic_set_led_brightness(self, value: int):
        pass

    async def optic_set_led_on_off(self, on: bool):
        pass

    async def optic_set_displacement_zero(self):
        pass

    async def optic_get_displacement_value(self):
        pass

    async def optic_detect_colony(
        self,
        image: cv2.Mat = None,
        resize_ratio: float = 0.5,
        padding: tuple[float, float, float, float] = (0.5, 0.5, 0.5, 0.5),
        rect_refer: cv2.typing.Rect = (8, 8, 83, 83),
        rect_tolerance: int = 5,
        vignette_size: int = 30,
        binary_thresh: int = 170,
        crop_border_thresh: int = 30,
        return_process_images: bool = False,
        font_scale: float = 0.7,
    ):
        pass

    async def optic_group_colony(
        self,
        crop_info: tuple[int, int, int, int],
        label_infos: list[LabelInfo],
        image: cv2.Mat = None,
        pick_count: int = 1,
        resize_ratio: float = 0.5,
        padding: tuple[float, float, float, float] = (0.5, 0.5, 0.5, 0.5),
        initial_dist: float = 60,
        row: int = 8,
        column: int = 12,
    ):
        pass



from dataclasses import dataclass
from enum import Enum
from typing import TypedDict

import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.common.notable.structs import Speed, FlowRate, LedBarParam
from ablelabs.neon.common.alma.enums import (
    LocationType,
    LocationReference,
    EquipmentStatus,
)


@dataclass
class Location:
    location_type: LocationType
    location_number: int = None
    well: str = None
    reference: LocationReference = None
    offset: tuple[float, float, float] = (0, 0, 0)

    def __str__(self):
        # 필드값이 기본값과 다를 때만 표시
        fields = []
        if self.location_number == None:
            fields.append(f"{self.location_type}")
        else:
            fields.append(f"{self.location_type}.{self.location_number}")
        if self.well != None:
            fields.append(f"well={self.well}")
        if self.reference != None:
            fields.append(f"reference={self.reference}")
        if self.offset != (0, 0, 0):
            fields.append(f"offset={self.offset}")
        return f"Location({', '.join(fields)})"


def location(
    location_type: LocationType = LocationType.DECK,
    location_number: int = None,
    well: str = None,
    reference: LocationReference = None,
    offset: tuple[float] = (0, 0, 0),
):
    return Location(
        location_type=location_type,
        location_number=location_number,
        well=well,
        reference=reference,
        offset=offset,
    )


class MeasurementTimeMode(Enum):
    SHORT = "SHOR"
    MEDIUM = "MED"
    LONG = "LONG"


@dataclass
class LCRParam:
    # preset value
    data_format_long: bool = False
    # function_impedance_type: MeasurementFunction = MeasurementFunction.CP_D
    frequency: float = 1000
    voltage_level: float = 1
    measurement_time_mode: MeasurementTimeMode = MeasurementTimeMode.MEDIUM
    averaging_rate: int = 1


class HeaterData(TypedDict):
    on: bool
    temperature: float


@dataclass
class State:
    equipment_status: EquipmentStatus = None
    location_type_number_well: tuple[LocationType, int, str] = None

    def to_dict(self):
        return {
            "equipment_status": self.equipment_status,
            "location_type_number_well": self.location_type_number_well,
        }

    @staticmethod
    def from_dict(data):
        data["equipment_status"] = EquipmentStatus(data["equipment_status"])
        data["location_type_number_well"] = (
            (
                LocationType(data["location_type_number_well"][0]),
                int(data["location_type_number_well"][1]),
                data["location_type_number_well"][2],
            )
            if data["location_type_number_well"]
            else None
        )
        return State(**data)

    def __str__(self):
        fields = []
        fields.append(f"{self.equipment_status}")
        if self.location_type_number_well != None:
            location_type, location_number, well = self.location_type_number_well
            location_str = location_type
            if location_number != None:
                location_str += f".{location_number}"
            if well != None:
                location_str += f"({well})"
            fields.append(f"{location_str}")
        return f"State({', '.join(fields)})"



# python 3.10 버전

# 가상환경 설정
## python -m venv .venv
## py 3.10 -m venv .venv

# ablelabs 패키지 설치
## pip install ablelabs

# Neon.exe + resources 파일 위치 확인
## resources 폴더는 Neon.exe와 같은 폴더 또는 하위 폴더에

# subprocess.Popen으로 Neon.exe 실행

# connect -> wait_boot -> set_pipettes/tips/labwares 순서대로 호출


import asyncio
import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.controllers.notable.api.robot_api import RobotAPI
from ablelabs.neon.common.notable.enums import (
    Axis,
    LocationReference,
    RunStatus,
)
from ablelabs.neon.common.notable.structs import Speed, FlowRate, location


async def main():
    import subprocess

    # for windows.
    # subprocess.Popen(
    #     [
    #         r"C:\Users\ABLE Labs\Desktop\ABLE-Elba\robot\dist\Neon.exe",
    #     ]
    # )

    # for mac.
    subprocess.Popen(
        [
            r"/Users/sypark/Code/ABLE-Elba/.venv/bin/python",
            r"/Users/sypark/Code/ABLE-Elba/robot/src/controllers/notable/v2/robot_router.py",
        ],
        cwd=r"/Users/sypark/Code/ABLE-Elba",
    )
    await asyncio.sleep(1)

    ip = "localhost"
    port = 1234

    robot_api = RobotAPI()
    try:
        await robot_api.connect(ip=ip, port=port)
    except Exception as e:
        pass

    await robot_api.wait_boot()
    await robot_api.set_run_status(RunStatus.READY)
    await robot_api.get_run_status()

    # set
    await robot_api.set.pipettes(
        {
            # 1: "8ch1000ul",
            2: "1ch200ul",
        }
    )
    await robot_api.set.tips(
        {
            # 1: "tip_1000",
            2: "tip_200",
        }
    )
    await robot_api.set.labwares(
        {
            1: "spl_trayplate_60ml_#30001",
            2: "spl_96_well_0.2ml_#30096",
            10: "ablelabs_tiprack_#AL-CT-200",
            11: "ablelabs_tiprack_#AL-CT-1000",
            12: "trash_#v2.5",
        }
    )

    # get
    await robot_api.get.pipette_infos()
    await robot_api.get.tip_infos()
    await robot_api.get.deck_module_infos()
    await robot_api.get.labware_infos()
    setup_data = await robot_api.get.setup_data()

    x, y, z = setup_data["teaching"]["pipette_1"]["8ch200ul"]["deck1_up"]
    setup_data["teaching"]["pipette_1"]["8ch200ul"]["deck1_up"] = [x + 1, y + 2, z + 3]
    await robot_api.set.update_setup_data(setup_data, True)

    # robot
    await robot_api.stop()
    await robot_api.clear_error()  # stop 이후, motion 전에.
    await robot_api.pause()
    await robot_api.resume()
    is_connected = await robot_api.is_connected()

    # motion
    await robot_api.motion.initialize()
    await robot_api.motion.move_to(
        pipette_number=2,
        location=location(
            location_number=1,
            well="a1",
            reference=LocationReference.BOTTOM,
            offset=(1.0, 2.0, 3.0),
        ),
    )

    await robot_api.motion.pick_up_tip(
        pipette_number=2,
        location=location(
            location_number=10,
            well="a1",
        ),
    )
    await robot_api.motion.drop_tip(
        pipette_number=2,
        location=location(
            location_number=10,
            well="a1",
        ),
    )
    await robot_api.motion.pick_up_tip(
        pipette_number=2,
        location=location(
            location_number=10,
            well="a12",
        ),
    )
    await robot_api.motion.drop_tip(
        pipette_number=2,
        location=location(
            location_number=12,
        ),
    )

    await robot_api.motion.aspirate(
        pipette_number=2,
        volume=200,
        location=location(
            location_number=1,
            well="a1",
            reference=LocationReference.BOTTOM,
        ),
        flow_rate=FlowRate.from_ul(100),
    )
    await robot_api.motion.rise_tip(
        pipette_number=2,
        height_offset=5,
        z_speed=Speed.from_mm(2),
    )
    await robot_api.motion.dispense(
        pipette_number=2,
        volume=200,
        location=location(
            location_number=2,
            well="a1",
            reference=LocationReference.BOTTOM,
        ),
        flow_rate=FlowRate.from_ul(100),
    )
    await robot_api.motion.mix(
        pipette_number=2,
        volume=100,
        iteration=2,
        # location=location(
        #     location_number=2,
        #     well="a1",
        #     reference=LocationReference.BOTTOM,
        # ),
        flow_rate=FlowRate.from_ul(70),
        delay=0.1,
    )
    await robot_api.motion.blow_out(
        pipette_number=2,
        flow_rate=FlowRate.from_ul(200),
    )
    await robot_api.motion.move_to_ready()

    # axis
    position = await robot_api.axis.get_position(axis=Axis.X)  # mm
    await robot_api.axis.set_speed(axis=Axis.X, value=10)  # mm/sec
    await robot_api.axis.set_accel(axis=Axis.X, value=10)  # mm/sec2
    await robot_api.axis.set_decel(axis=Axis.X, value=10)  # mm/sec2
    await robot_api.axis.disable(axis=Axis.X)
    await robot_api.axis.enable(axis=Axis.X)
    await robot_api.axis.stop(axis=Axis.X)
    await robot_api.axis.home(axis=Axis.X)
    await robot_api.axis.wait_home_done(axis=Axis.X)
    await robot_api.axis.jog(axis=Axis.X, value=10)  # mm/sec
    await robot_api.axis.jog(axis=Axis.X, value=0)  # mm/sec
    await robot_api.axis.step(axis=Axis.X, value=10)  # mm
    await robot_api.axis.wait_move_done(axis=Axis.X)
    await robot_api.axis.move(axis=Axis.X, value=10)  # mm
    await robot_api.axis.wait_move_done(axis=Axis.X)


if __name__ == "__main__":
    loop = asyncio.new_event_loop()
    loop.create_task(main())
    loop.run_forever()



import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.network.messenger import MessengerServer
from ablelabs.neon.common.notable.nanophilia.enums import (
    Axis,
    AxisDO,
    DioPDO,
    DioInput,
    RunStatus,
)
from ablelabs.neon.common.notable.nanophilia.structs import Location, Speed, FlowRate


class RobotRouter(MessengerServer):
    # robot
    async def robot_wait_boot(self):
        pass

    async def robot_shutdown(self):
        pass

    async def robot_stop(self):
        pass

    async def robot_clear_error(self):
        pass

    async def robot_pause(self):
        pass

    async def robot_resume(self):
        pass

    async def robot_is_connected(self) -> dict[str, bool]:
        pass

    async def robot_get_dio_input(self) -> dict[DioInput, bool]:
        pass

    async def robot_get_environment(self) -> dict[str, float]:
        pass

    async def robot_get_run_status(self) -> RunStatus:
        pass

    async def robot_get_changed_robot_status(self) -> dict:
        pass

    async def robot_set_interlock(self, value: bool):
        pass

    async def robot_set_led_lamp(self, on: bool):
        pass

    async def robot_set_run_status(self, value: RunStatus):
        pass

    async def robot_set_progress_rate(self, value: float):
        pass

    async def robot_initialize(self):
        pass

    async def robot_delay(self, sec: float):
        pass

    # ffu
    async def robot_get_ffu(self) -> dict:
        pass

    async def robot_set_ffu(self, on: bool, rpm: float = None):
        pass

    async def robot_set_ffu_alarm_clear(self):
        pass

    # get
    async def get_pipette_infos(self) -> dict[int, dict]:
        pass

    async def get_tip_infos(self) -> dict[int, dict]:
        pass

    async def get_deck_module_infos(self) -> dict[int, dict]:
        pass

    async def get_labware_infos(self) -> dict[int, dict]:
        pass

    async def get_setup_data(self) -> dict:
        pass

    async def get_region_dataset(self) -> dict:
        pass

    async def get_pipette_dataset(self) -> dict:
        pass

    async def get_tip_dataset(self) -> dict:
        pass

    async def get_deck_module_dataset(self) -> dict:
        pass

    async def get_labware_dataset(self) -> dict:
        pass

    async def get_region_json(self) -> dict:
        pass

    async def get_pipette_json(self) -> dict:
        pass

    async def get_tip_json(self) -> dict:
        pass

    async def get_deck_module_json(self) -> dict:
        pass

    async def get_labware_json(self) -> dict:
        pass

    async def get_setup_data_toml(self) -> str:
        pass

    async def get_driver_param_toml(self) -> str:
        pass

    # set
    async def set_pipettes(self, value: dict[int, str | None]):
        pass

    async def set_tips(self, value: dict[int, str | None]):
        pass

    async def set_deck_modules(self, value: dict[int, str | None]):
        pass

    async def set_labwares(self, value: dict[int, str | None]):
        pass

    async def set_update_setup_data(self, value: dict, save: bool):
        pass

    async def set_update_pipette_attrs(self, value: dict[int, dict]):
        pass

    # motion(pipette)
    async def motion_initialize(self):
        pass

    async def motion_move_z_up(self):
        pass

    async def motion_move_to_ready(self):
        pass

    async def motion_move_to(
        self,
        pipette_number: int,
        location: Location,
        optimize: bool = False,
    ):
        pass

    async def motion_pick_up_tip(
        self,
        pipette_number: int,
        location: Location,
        optimize: bool = False,
    ):
        pass

    async def motion_drop_tip(
        self,
        pipette_number: int,
        location: Location,
        optimize: bool = False,
    ):
        pass

    async def motion_rise_tip(
        self,
        pipette_number: int,
        height_offset: float,
        z_speed: Speed,
    ):
        pass

    async def motion_ready_plunger(
        self,
        pipette_number: int,
        flow_rate: FlowRate = None,
    ):
        pass

    async def motion_blow_out(
        self,
        pipette_number: int,
        flow_rate: FlowRate = None,
    ):
        pass

    async def motion_aspirate(
        self,
        pipette_number: int,
        volume: float,
        location: Location = None,
        flow_rate: FlowRate = None,
        optimize: bool = False,
    ):
        pass

    async def motion_dispense(
        self,
        pipette_number: int,
        volume: float,
        location: Location = None,
        flow_rate: FlowRate = None,
        optimize: bool = False,
    ):
        pass

    async def motion_mix(
        self,
        pipette_number: int,
        volume: float,
        iteration: int,
        location: Location = None,
        flow_rate: FlowRate = None,
        delay: float = 0.0,
        optimize: bool = False,
    ):
        pass

    # deck_module.washer_dryer
    async def deck_module_washer_dryer_off_do(self):
        pass

    async def deck_module_washer_dryer_initialize(self):
        pass

    async def deck_module_washer_dryer_move_to_ready(self):
        pass

    async def deck_module_washer_dryer_move_to_washing(self):
        pass

    async def deck_module_washer_dryer_move_to(
        self,
        column: int,
        offset: tuple[float, float, float] = (0, 0, 0),
    ):
        pass

    async def deck_module_washer_dryer_prime(self, sec: float):
        pass

    async def deck_module_washer_dryer_recovery(self, sec: float):
        pass

    async def deck_module_washer_dryer_suction(
        self,
        columns: list[int],
        depth: float,
        z_speed: Speed = None,
        delay_sec: float = 0,
    ):
        pass

    async def deck_module_washer_dryer_dispense(
        self,
        columns: list[int],
        volume: float,
    ):
        pass

    async def deck_module_washer_dryer_wash_needle(
        self,
        depth: float,
        delay_sec: float,
    ):
        pass

    async def deck_module_washer_dryer_wash_tube(
        self,
        cycle: int,
        suction_sec: float,
        dispense_sec: float,
    ):
        pass

    async def deck_module_washer_dryer_dry(self, sec: float):
        pass

    # deck_module.magnetic_shaker
    async def deck_module_magnetic_shaker_initialize(self):
        pass

    async def deck_module_magnetic_shaker_move_to_ready(self):
        pass

    async def deck_module_magnetic_shaker_shake(self, rpm: int, acceleration_sec: int):
        pass

    async def deck_module_magnetic_shaker_shake_off(self):
        pass

    async def deck_module_magnetic_shaker_magnet(self, on: bool):
        pass

    # deck_module.shaker
    async def deck_module_shaker_initialize(self):
        pass

    async def deck_module_shaker_move_to_ready(self):
        pass

    async def deck_module_shaker_shake(self, rpm: int, acceleration_sec: int):
        pass

    async def deck_module_shaker_shake_off(self):
        pass

    # axis
    async def axis_get_position(self, axis: Axis, floor_digit: int = 1):
        pass

    async def axis_set_speed(self, axis: Axis, value: float):
        pass

    async def axis_set_accel(self, axis: Axis, value: float):
        pass

    async def axis_set_decel(self, axis: Axis, value: float):
        pass

    async def axis_enable(self, axis: Axis):
        pass

    async def axis_disable(self, axis: Axis):
        pass

    async def axis_stop(self, axis: Axis):
        pass

    async def axis_home(self, axis: Axis):
        pass

    async def axis_jog(self, axis: Axis, value: float):
        pass

    async def axis_step(self, axis: Axis, value: float):
        pass

    async def axis_move(self, axis: Axis, value: float):
        pass

    async def axis_wait_home_done(self, axis: Axis):
        pass

    async def axis_wait_move_done(self, axis: Axis):
        pass



from openpyxl import load_workbook
from openpyxl.worksheet.worksheet import Worksheet
from openpyxl.utils.cell import get_column_letter
from openpyxl.cell.cell import Cell
from openpyxl.worksheet.formula import ArrayFormula
from typing import Optional


class ExcelUtils:
    def __init__(self, path: str, data_only: bool = False) -> None:
        """
        :param path: path to Excel file
        :param data_only: if any formula exists in workbook,
            read operation will return calculated value of each cell.
            else, read operation will return the formula itself.
            (openpyxl's default value is False.)
        """
        self._path = path
        self._workbook = load_workbook(filename=path, data_only=data_only)
        self._worksheet: Worksheet = None

    def save(self):
        self._workbook.save(filename=self._path)

    def select_sheet(self, sheet_name: str):
        self._worksheet = self._workbook[sheet_name]

    def get_cell(self, address: str):
        cell: Cell = self._worksheet[address]
        return cell.value

    def get_cell_by_row_column(self, row: int, column: int):
        cell: Cell = self._worksheet.cell(row=row, column=column)
        return cell.value

    def get_range(self, address1: str, address2: str):
        cells: tuple[tuple[Cell]] = self._worksheet[address1:address2]
        result = [[cell.value for cell in row] for row in cells]
        return result

    def get_range_by_rows(self):
        result = [[cell.value for cell in row] for row in self._worksheet.rows]
        return result

    def set_cell(self, address: str, value, array_formula: bool = False, array_ref: Optional[str] = None):
        if array_formula:
            value = ArrayFormula(ref=address if array_ref is None else array_ref, text=value)
        self._worksheet[address] = value

    def set_cell_by_row_column(self, row: int, column: int, value):
        self._worksheet.cell(row=row, column=column, value=value)

    def append_row(self, values: list | dict):
        self._worksheet.append(values)

    def add_filter(self, row_size: int, column_size: int):
        """
        Add whole-range filter in active worksheet.
        Asserts data starts from Cell A1.
        :param row_size: number of rows (int, >0)
        :param column_size: number of columns (int, >0)
        """
        # 240610 Dongha Kim - to add filter in 'calculate' and 'raw data'
        def number_to_coordinate(row_idx, col_idx):
            assert all([
                isinstance(row_idx, int),
                isinstance(col_idx, int),
                row_idx > 0,
                col_idx > 0
            ])
            col_string = get_column_letter(col_idx)
            return f'{col_string}{row_idx}'
        start = number_to_coordinate(1, 1)
        end = number_to_coordinate(row_size, column_size)
        self._worksheet.auto_filter.ref = f'{start}:{end}'
        self._worksheet.auto_filter.add_filter_column(0, [])


if __name__ == "__main__":
    import pathlib
    import shutil

    absolute_dir = pathlib.Path(__file__).resolve()  # ABLE-Elba\robot\src\utils\excel_utils.py
    root = absolute_dir.parent.parent.parent.parent
    dir_path = root / "robot" / "resources" / "platform_autocal" / "template"
    source_file = f"{dir_path}/template_(1)_cal.xlsx"
    target_file = f"{dir_path}/result_(1)_cal.xlsx"
    shutil.copyfile(source_file, target_file)

    excel = ExcelUtils(target_file)
    excel.select_sheet("raw data")

    # cell = excel.get_cell("A2")
    # print(cell)
    # cell_range = excel.get_range("A1", "K1")
    # print(cell_range)
    # cell_range = excel.get_range_by_rows()
    # print(cell_range)

    volumes = [200, 100, 20]

    # 엑셀에 raw data 쓰기.
    import datetime
    import random

    index = 1
    reuse_tip = 1
    reuse_tip_count = 5
    tip_position_index = 1
    tip_position = f"A{tip_position_index}"
    for volume in volumes:
        for channel in range(1, 9):
            for repeat in range(1, 6):
                weight = volume * (1 + 1 * (0.5 - random.random()))
                pressure = 1000 + 5 * (0.5 - random.random())
                humidity = 50 + 1 * (0.5 - random.random())
                temperature = 20 + 0.2 * (0.5 - random.random())
                values = [
                    index,
                    datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                    channel,
                    repeat,
                    reuse_tip,
                    tip_position,
                    volume,
                    weight,
                    pressure,
                    humidity,
                    temperature,
                ]
                excel.append_row(values)
                # excel.save()
                print(values)

                index += 1
                reuse_tip += 1
                if reuse_tip > reuse_tip_count:
                    reuse_tip = 1
                    tip_position_index += 1
                    tip_position = f"A{tip_position_index}"
    # A ~ K 열, 1 ~ 121 행 Filter 추가
    excel.add_filter(row_size=121, column_size=11)
    excel.save()

    # 엑셀 읽어서 calculate 계산.
    import pandas as pd
    from pandas import DataFrame
    import numpy as np

    raw_data = pd.read_excel(target_file, sheet_name="raw data")
    len_raw_data = len(raw_data)
    print(raw_data)

    volume_channel_weights = {
        volume: raw_data.loc[
            raw_data["volume (μL)"] == volume, ["channel", "weight (mg)"]
        ]
        for volume in volumes
    }
    volume_calculates: dict[float, DataFrame] = {}
    for volume, channel_weights in volume_channel_weights.items():
        print(f"\nvolume={volume}")
        print(channel_weights)
        grouped_weights = channel_weights.groupby("channel")
        calculate = grouped_weights.agg(
            [
                "mean",
                # ddof=0: 모집단 표준편차(STDEV.P) / ddof=1(default): 표본 표준편차(STDEV.S)
                lambda x: np.std(x, ddof=0),
                lambda x: np.std(x) / np.mean(x),
                lambda x: (volume - np.mean(x)) / volume,
            ]
        )
        calculate.columns = ["avg", "std", "cv", "acc"]
        max_cv_idx = calculate["cv"].idxmax()
        max_acc_idx = calculate["acc"].abs().idxmax()
        calculate["max cv"] = ""
        calculate["max acc"] = ""
        calculate.loc[max_cv_idx, "max cv"] = calculate.loc[max_cv_idx, "cv"]
        calculate.loc[max_acc_idx, "max acc"] = calculate.loc[max_acc_idx, "acc"]
        calculate = calculate.reset_index()
        calculate.insert(0, "volume", volume)
        print(calculate)
        volume_calculates[volume] = calculate

    # 엑셀에 calculate 쓰기.
    # df.to_excel()은 새로 쓰기 때문에 기존 데이터가 제거됨.
    # for calculate in volume_calculates.values():
    #     calculate.to_excel(
    #         target_file,
    #         sheet_name="calculate",
    #         startcol=ord("J") - ord("A"),
    #         startrow=1,
    #     )

    excel.select_sheet("calculate")
    r = 2
    for calculate in volume_calculates.values():
        for i, row in calculate.iterrows():
            (volume, channel, avg, std, cv, acc, max_cv, max_acc) = row
            excel.select_sheet("calculate")
            excel.set_cell(f"A{r}", volume)
            excel.set_cell(f"B{r}", channel)
            excel.set_cell(f"C{r}", avg)
            excel.set_cell(f"D{r}", std)
            excel.set_cell(f"E{r}", cv)
            excel.set_cell(f"F{r}", acc)
            excel.set_cell(f"G{r}", max_cv)
            excel.set_cell(f"H{r}", max_acc)
            excel.select_sheet("calculate_f")
            excel.set_cell(f"A{r}", volume)
            excel.set_cell(f"B{r}", channel)
            r += 1
    r -= 1
    excel.select_sheet("calculate_f")
    for rr in range(2, r + 1):
        excel.set_cell(
            f"C{rr}",
            f"=AVERAGEIFS("
            f"'raw data'!$H$2:$H${len_raw_data + 1}, "
            f"'raw data'!$C$2:$C${len_raw_data + 1}, $B{rr}, 'raw data'!$G$2:$G${len_raw_data + 1}, $A{rr}"
            f")",
            array_formula=True
        )
        excel.set_cell(
            f"D{rr}",
            f"=_xlfn.STDEV.P(IF(("
            f"'raw data'!$C$2:$C${len_raw_data + 1}=$B{rr})*('raw data'!$G$2:$G${len_raw_data + 1}=$A{rr}), "
            f"'raw data'!$H$2:$H${len_raw_data + 1}"
            f"))",
            array_formula=True
        )
        excel.set_cell(f"E{rr}", f"=$D{rr}/$C{rr}")
        excel.set_cell(f"F{rr}", f"=($A{rr}-$C{rr})/$A{rr}")
        excel.set_cell(
            f"G{rr}",
            f"=IF(ABS($E{rr})=MAX(IF($A$2:$A${r}=$A{rr}, ABS($E$2:$E${r}))), ABS($E{rr}), \"\")",
            array_formula=True
        )
        excel.set_cell(
            f"H{rr}",
            f"=IF(ABS($F{rr})=MAX(IF($A$2:$A${r}=$A{rr}, ABS($F$2:$F${r}))), ABS($F{rr}), \"\")",
            array_formula=True
        )
    for sheet_name in ("calculate_f", "calculate"):
        excel.select_sheet(sheet_name)
        excel.add_filter(r, 8)
    excel.save()

    # calibartion 계산.
    from collections import defaultdict

    cal_x = []
    cal_y = []
    for volume, calculate in volume_calculates.items():
        avg_weight_of_volume = calculate["avg"].mean()
        cal_x.append(avg_weight_of_volume)  # 측정값
        cal_y.append(volume)  # 목표값
    _cal_x = [int(x * 10) / 10 for x in cal_x]
    print(f"cal_x={_cal_x}")
    print(f"cal_y={cal_y}")

    from sklearn.linear_model import LinearRegression

    reg = LinearRegression()
    reg_X = [[x] for x in cal_x]
    reg_y = cal_y
    reg.fit(X=reg_X, y=reg_y)
    r2 = reg.score(X=reg_X, y=reg_y)
    a = reg.coef_[0]
    b = reg.intercept_
    print(f"a={a:.4f}  b={b:.4f}  r2={r2:.4f}")

    # 엑셀에 calibartion 쓰기.
    for i in range(len(cal_x)):
        rr = 7 + i
        excel.select_sheet("calculate")
        excel.set_cell(f"J{rr}", cal_x[i])
        excel.set_cell(f"K{rr}", cal_y[i])
        excel.select_sheet("calculate_f")
        excel.set_cell(f"K{rr}", cal_y[i])
        excel.set_cell(
            f"J{rr}",
            f"=AVERAGEIFS($C$2:$C${r}, $A$2:$A${r}, $K{rr})",
            array_formula=True
        )
    excel.select_sheet("calculate")
    excel.set_cell("K2", a)
    excel.set_cell("K3", b)
    excel.set_cell("K4", r2)
    fmt = {"end_row": 6 + len(reg_X)}
    excel.set_cell(
        "L2",
        r"=INDEX(LINEST($K$7:$K$%(end_row)d,$J$7:$J$%(end_row)d^{1},1,0),1)" % fmt,
        array_formula=True,
    )
    excel.set_cell(
        "L3",
        r"=INDEX(LINEST($K$7:$K$%(end_row)d,$J$7:$J$%(end_row)d^{1},1,0),2)" % fmt,
        array_formula=True,
    )
    excel.set_cell(
        "L4",
        r"=INDEX(LINEST($K$7:$K$%(end_row)d,$J$7:$J$%(end_row)d^{1},1,1),3,1)" % fmt,
        array_formula=True,
    )
    excel.select_sheet("calculate_f")
    excel.set_cell(
        "K2",
        r"=INDEX(LINEST($K$7:$K$%(end_row)d,$J$7:$J$%(end_row)d^{1},1,0),1)" % fmt,
    )
    excel.set_cell(
        "K3",
        r"=INDEX(LINEST($K$7:$K$%(end_row)d,$J$7:$J$%(end_row)d^{1},1,0),2)" % fmt,
        array_formula=True
    )
    excel.set_cell(
        "K4",
        r"=INDEX(LINEST($K$7:$K$%(end_row)d,$J$7:$J$%(end_row)d^{1},1,1),3,1)" % fmt,
        array_formula=True
    )
    excel.set_cell("L2", None)
    excel.set_cell("L3", None)
    excel.set_cell("L4", None)
    excel.save()



import asyncio
from loguru import logger

import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.network.messenger import MessengerClient, run_server_func
from ablelabs.neon.utils.network.tcp_client import TcpClient
from ablelabs.neon.controllers.suitable.sara.api.robot_router import RobotRouter
from ablelabs.neon.controllers.suitable.sara.api.get_api import GetAPI
from ablelabs.neon.controllers.suitable.sara.api.set_api import SetAPI
from ablelabs.neon.controllers.suitable.sara.api.time_api import TimeAPI
from ablelabs.neon.controllers.suitable.sara.api.motion_api import MotionAPI
from ablelabs.neon.controllers.suitable.sara.api.axis_api import AxisAPI
from ablelabs.neon.common.suitable.constants import PIPETTE_NUMBERS
from ablelabs.neon.common.suitable.enums import (
    LocationType,
    Axis,
    LocationReference,
    RunStatus,
    PipetteCalibrationType,
)
from ablelabs.neon.common.suitable.structs import Speed, FlowRate, location


class RobotAPI(MessengerClient):
    def __init__(self) -> None:
        tcp_client = TcpClient(name="tcp_client", log_func=logger.trace)
        super().__init__(tcp_client)
        self._get_api = GetAPI(tcp_client=tcp_client)
        self._set_api = SetAPI(tcp_client=tcp_client)
        self._time_api = TimeAPI(tcp_client=tcp_client)
        self._motion_api = MotionAPI(tcp_client=tcp_client)
        self._axis_api = AxisAPI(tcp_client=tcp_client)

    @property
    def get(self):
        return self._get_api

    @property
    def set(self):
        return self._set_api

    @property
    def time(self):
        return self._time_api

    @property
    def motion(self):
        return self._motion_api

    @property
    def axis(self):
        return self._axis_api

    async def connect(self, ip, port):
        await self._tcp_client.connect(ip=ip, port=port)

    @run_server_func(RobotRouter.robot_wait_boot)
    async def wait_boot(self):
        pass

    @run_server_func(RobotRouter.robot_stop)
    async def stop(self):
        pass

    @run_server_func(RobotRouter.robot_clear_error)
    async def clear_error(self):
        pass

    @run_server_func(RobotRouter.robot_pause)
    async def pause(self):
        pass

    @run_server_func(RobotRouter.robot_resume)
    async def resume(self):
        pass

    @run_server_func(RobotRouter.robot_is_connected)
    async def is_connected(self):
        pass

    @run_server_func(RobotRouter.robot_get_run_status)
    async def get_run_status(self) -> dict[str, RunStatus | bool]:
        pass

    @run_server_func(RobotRouter.robot_get_changed_robot_status)
    async def get_changed_robot_status(self) -> dict[str, RunStatus | bool]:
        pass

    @run_server_func(RobotRouter.robot_set_run_status)
    async def set_run_status(self, value: RunStatus):
        pass

    @run_server_func(RobotRouter.robot_set_progress_rate)
    async def set_progress_rate(self, value: float):
        pass

    @run_server_func(RobotRouter.robot_set_interlock)
    async def set_interlock(self, value: bool):
        pass


async def main():
    import subprocess

    # for windows.
    # root = r"D:\Code\ABLE-Elba"
    # subprocess.Popen(
    #     [
    #         rf"{root}\.venv\Scripts\python.exe",
    #         rf"{root}\robot\src\controllers\suitable\sara\robot_router.py",
    #     ],
    #     cwd=root,
    # )
    # await asyncio.sleep(1)

    # for mac.
    # subprocess.Popen(
    #     [
    #         r"/Users/sypark/Code/ABLE-Elba/.venv/bin/python",
    #         r"/Users/sypark/Code/ABLE-Elba/robot/src/controllers/suitable/sara/robot_router.py",
    #     ],
    #     cwd=r"/Users/sypark/Code/ABLE-Elba",
    # )
    # await asyncio.sleep(1)

    logger.remove()
    # logger.add(sys.stdout, level="TRACE")
    # logger.add(sys.stdout, level="DEBUG")
    logger.add(sys.stdout, level="INFO")
    # logger.add("logs/trace.log", level="TRACE")
    # logger.add("logs/debug.log", level="DEBUG")
    logger.add("logs/info.log", level="INFO")

    ip = "localhost"
    port = 1234

    robot_api = RobotAPI()
    try:
        await robot_api.connect(ip=ip, port=port)
    except Exception as e:
        pass

    # get
    logger.info(f"setup_data = {await robot_api.get.setup_data()}")

    # set
    await robot_api.set.pipettes({n: "1ch1000ul" for n in PIPETTE_NUMBERS})
    await robot_api.set.tips({n: "tip_1000" for n in PIPETTE_NUMBERS})
    await robot_api.set.labwares(
        {
            21: "ablelabs_tiprack_200ul",
            22: "ablelabs_tiprack_200ul",
            23: "ablelabs_tiprack_200ul",
            24: "ablelabs_tiprack_200ul",
            4: "ablelabs_tiprack_1000ul",
            6: "ablelabs_tiprack_1000ul",
            # 5: "spl_dw_reservoir",
            5: "ablelabs_acn_reservoir",
            15: "bioneer_96_qc_plate",
            16: "bioneer_96_qc_plate",
            9: "bioneer_96_qc_plate",
            # 10: "bioneer_96_qc_plate",
            10: "bioneer_384_qc_plate",
            13: "spl_96_deep_well_plate",
            14: "spl_96_deep_well_plate",
            7: "spl_96_deep_well_plate",
            8: "spl_96_deep_well_plate",
            17: "spl_96_deep_well_plate",
            18: "spl_96_deep_well_plate",
            11: "spl_96_deep_well_plate",
            12: "spl_96_deep_well_plate",
        }
    )
    await robot_api.set.update_pipette_attrs(
        {n: {"blow_out_volume": 20} for n in PIPETTE_NUMBERS}
    )
    await robot_api.set.pipette_calibrations(
        {n: PipetteCalibrationType.Tip_1000_DW for n in PIPETTE_NUMBERS}
    )

    # robot
    await robot_api.stop()
    await robot_api.clear_error()  # stop 이후, motion 전에. / fault 후에.
    await robot_api.pause()
    await robot_api.resume()
    logger.info(f"is_connected = {await robot_api.is_connected()}")
    await robot_api.set_run_status(RunStatus.INITIALIZE)
    await robot_api.set_run_status(RunStatus.RUN)
    await robot_api.set_progress_rate(5)
    await robot_api.set_interlock(False)
    logger.info(f"get_run_status = {await robot_api.get_run_status()}")

    async def get_changed_robot_status():
        while True:
            robot_status = await robot_api.get_changed_robot_status()
            logger.info(f"robot_status = {robot_status}")

    asyncio.create_task(get_changed_robot_status())
    await asyncio.sleep(0.5)
    await robot_api.set_run_status(RunStatus.DONE)

    # time
    await robot_api.time.initialize()
    await robot_api.time.move_to_ready()
    await robot_api.time.move_to(
        pipette_number=[1, 2, 3],
        location=location(
            location_type=LocationType.DECK,
            location_number=[4, 4, 4],
            well=["a1", "b1", "c1"],
        ),
    )

    await robot_api.time.pick_up_tip(
        pipette_number=[1, 2, 3, 4, 5, 6, 7, 8],
        location=location(
            location_type=LocationType.DECK,
            location_number=[3, 3, 3, 3, 3, 3, 3, 3],
            well=["a1", "b1", "c1", "d1", "e1", "f1", "g1", "h1"],
        ),
    )
    await robot_api.time.drop_tip(
        pipette_number=[1, 2, 3, 4, 5, 6, 7, 8],
        location=location(
            location_type=LocationType.DECK,
            location_number=[3, 3, 3, 3, 3, 3, 3, 3],
            well=["a1", "b1", "c1", "d1", "e1", "f1", "g1", "h1"],
        ),
    )
    await robot_api.time.drop_tip(
        pipette_number=[1, 2, 3, 4, 5, 6, 7, 8],
        location=location(
            location_type=LocationType.WASTE,
        ),
    )

    await robot_api.time.aspirate(
        pipette_number=[1, 2, 3, 4, 5, 6, 7, 8],
        volume=200,
        location=location(
            location_type=LocationType.DECK,
            location_number=[2, 2, 2, 2, 2, 2, 2, 2],
            well=["a1", "b1", "c1", "d1", "e1", "f1", "g1", "h1"],
            reference=LocationReference.LIQUID,
        ),
        flow_rate=FlowRate.from_ul(100),
        rise_tip_height_offset=5,
        pre_wet_count=3,
    )
    await robot_api.time.rise_tip(
        pipette_number=[1, 2, 3, 4, 5, 6, 7, 8],
        height_offset=5,
        z_speed=Speed.from_mm(2),
    )
    await robot_api.time.dispense(
        pipette_number=[1, 2, 3, 4, 5, 6, 7, 8],
        volume=200,
        location=location(
            location_type=LocationType.DECK,
            location_number=[4, 4, 4, 4, 4, 4, 4, 4],
            well=["a1", "b1", "c1", "d1", "e1", "f1", "g1", "h1"],
            reference=LocationReference.BOTTOM,
        ),
        flow_rate=FlowRate.from_ul(100),
    )
    await robot_api.time.mix(
        pipette_number=[1, 2, 3, 4, 5, 6, 7, 8],
        volume=100,
        iteration=2,
        # location=location(
        #     location_type=LocationType.DECK,
        #     location_number=[4, 4, 4, 4, 4, 4, 4, 4],
        #     well=["a1", "b1", "c1", "d1", "e1", "f1", "g1", "h1"],
        #     reference=LocationReference.BOTTOM,
        # ),
        flow_rate=FlowRate.from_ul(70),
        delay=0.1,
    )
    await robot_api.time.blow_out(
        pipette_number=[1, 2, 3, 4, 5, 6, 7, 8],
        flow_rate=FlowRate.from_ul(200),
    )

    # motion
    await robot_api.motion.initialize()
    await robot_api.motion.home_z()
    await robot_api.motion.home_x()
    await robot_api.motion.home_y()
    await robot_api.motion.home_p()

    await robot_api.set.tips({n: "tip_1000" for n in PIPETTE_NUMBERS})
    await robot_api.motion.move_to(
        pipette_number=[1, 2, 3],
        location=location(
            location_type=LocationType.DECK,
            location_number=[4, 4, 4],
            well=["a1", "b1", "c1"],
        ),
    )

    await robot_api.motion.pick_up_tip(
        pipette_number=[1, 2, 3, 4, 5, 6, 7, 8],
        location=location(
            location_type=LocationType.DECK,
            location_number=[3, 3, 3, 3, 3, 3, 3, 3],
            well=["a1", "b1", "c1", "d1", "e1", "f1", "g1", "h1"],
        ),
    )
    await robot_api.motion.drop_tip(
        pipette_number=[1, 2, 3, 4, 5, 6, 7, 8],
        location=location(
            location_type=LocationType.DECK,
            location_number=[3, 3, 3, 3, 3, 3, 3, 3],
            well=["a1", "b1", "c1", "d1", "e1", "f1", "g1", "h1"],
        ),
    )
    await robot_api.motion.drop_tip(
        pipette_number=[1, 2, 3, 4, 5, 6, 7, 8],
        location=location(
            location_type=LocationType.WASTE,
        ),
    )

    await robot_api.motion.aspirate(
        pipette_number=[1, 2, 3, 4, 5, 6, 7, 8],
        volume=200,
        location=location(
            location_type=LocationType.DECK,
            location_number=[2, 2, 2, 2, 2, 2, 2, 2],
            well=["a1", "b1", "c1", "d1", "e1", "f1", "g1", "h1"],
            reference=LocationReference.LIQUID,
        ),
        flow_rate=FlowRate.from_ul(100),
    )
    await robot_api.motion.rise_tip(
        pipette_number=[1, 2, 3, 4, 5, 6, 7, 8],
        height_offset=5,
        z_speed=Speed.from_mm(2),
    )
    await robot_api.motion.dispense(
        pipette_number=[1, 2, 3, 4, 5, 6, 7, 8],
        volume=200,
        location=location(
            location_type=LocationType.DECK,
            location_number=[4, 4, 4, 4, 4, 4, 4, 4],
            well=["a1", "b1", "c1", "d1", "e1", "f1", "g1", "h1"],
            reference=LocationReference.BOTTOM,
        ),
        flow_rate=FlowRate.from_ul(100),
    )
    await robot_api.motion.mix(
        pipette_number=[1, 2, 3, 4, 5, 6, 7, 8],
        volume=100,
        iteration=2,
        # location=location(
        #     location_type=LocationType.DECK,
        #     location_number=[4, 4, 4, 4, 4, 4, 4, 4],
        #     well=["a1", "b1", "c1", "d1", "e1", "f1", "g1", "h1"],
        #     reference=LocationReference.BOTTOM,
        # ),
        flow_rate=FlowRate.from_ul(70),
        delay=0.1,
    )
    await robot_api.motion.blow_out(
        pipette_number=[1, 2, 3, 4, 5, 6, 7, 8],
        flow_rate=FlowRate.from_ul(200),
    )

    await robot_api.motion.aspirate(
        pipette_number=[1, 2, 3],
        volume=[[200, 100, 50]],
        location=location(
            location_type=LocationType.DECK,
            location_number=[2, 2, 2],
            well=["a1", "b1", "c1"],
            reference=LocationReference.BOTTOM,
        ),
        flow_rate=[FlowRate.from_ul(100), FlowRate.from_ul(80), FlowRate.from_ul(60)],
    )
    await robot_api.motion.blow_out(
        pipette_number=[1, 2, 3, 4, 5, 6, 7, 8],
        flow_rate=FlowRate.from_ul(200),
    )
    await robot_api.motion.move_to_ready()

    # axis
    position = await robot_api.axis.get_position(axis=Axis.X)  # mm
    await robot_api.axis.set_speed(axis=Axis.X, value=10)  # mm/sec
    await robot_api.axis.set_accel(axis=Axis.X, value=10)  # mm/sec2
    await robot_api.axis.set_decel(axis=Axis.X, value=10)  # mm/sec2
    await robot_api.axis.disable(axis=Axis.X)
    await robot_api.axis.enable(axis=Axis.X)
    await robot_api.axis.stop(axis=Axis.X)
    await robot_api.axis.home(axis=Axis.X)
    await robot_api.axis.wait_home_done(axis=Axis.X)
    await robot_api.axis.jog(axis=Axis.X, value=10)  # mm/sec
    await robot_api.axis.step(axis=Axis.X, value=10)  # mm
    await robot_api.axis.wait_move_done(axis=Axis.X)
    await robot_api.axis.move(axis=Axis.X, value=10)  # mm
    await robot_api.axis.wait_move_done(axis=Axis.X)


if __name__ == "__main__":
    asyncio.run(main())



DECK_COUNT: int = 12
DECK_NUMBERS = [i + 1 for i in range(DECK_COUNT)]

LID_HOLDER_COUNT: int = 4
LID_HOLDER_NUMBERS = [i + 1 for i in range(LID_HOLDER_COUNT)]

CHIP_HOLDER_COUNT: int = 8
CHIP_HOLDER_NUMBERS = [i + 1 for i in range(CHIP_HOLDER_COUNT)]

RESERVOIR_COUNT: int = 2
RESERVOIR_NUMBERS = [i + 1 for i in range(RESERVOIR_COUNT)]

CHIP_WELL_COUNT: int = 192



import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.common.notable.enums import (
    NUMBERING_ORDER,
    Color,
    RunStatus,
    TaskStatus,
    LabwareType,
    Height,
    Plunger,
    LocationReference,
)
from ablelabs.neon.utils.enum_types import Enum, auto, StrEnum, ItemGetable


class DioInput(StrEnum):
    EMS = auto()
    DOOR_LEFT = auto()
    DOOR_RIGHT = auto()


class DioPDO(StrEnum):
    LED_LAMP = auto()


class Axis(StrEnum):
    X = auto()
    Y = auto()
    Z1 = auto()
    Z2 = auto()
    G = auto()
    P = auto()

    Z = ItemGetable[int](lambda number: Axis.from_str(f"{Axis.Z1[0]}{number}"))


class LocationType(StrEnum):
    READY = auto()
    LID_HOLDER = auto()
    BLOCK_HOLDER = auto()
    INSPECTOR = auto()
    DECK = auto()
    CHIP_HOLDER = auto()
    WASTE = auto()
    TIP_RACK = auto()
    RESERVOIR = auto()


class ZGripper(StrEnum):
    MOVE = auto()
    PLATE_LID_UNGRIP = auto()
    PLATE_LID_GRIP = auto()
    LID_UNGRIP = auto()
    LID_GRIP = auto()
    BLOCK_PUSH = auto()
    BLOCK_GRIP = auto()
    FORK = auto()
    UNDER_PLATE = auto()
    PUSH = auto()


class Gripper(StrEnum):
    UNGRIP = auto()
    PUSH = auto()
    GRIP_LID = auto()
    FORK = auto()
    GRIP_PLATE = auto()
    GRIP_BLOCK = auto()
    PUSH_BLOCK = auto()


class EquipmentStatus(StrEnum):
    NORMAL = auto()
    BLOCK_FORK = auto()
    BLOCK_GRIP = auto()
    CHIP_FORK = auto()
    CHIP_GRIP = auto()
    LID_GRIP = auto()
    INSPECTING = auto()
    PICKED_UP_TIP = auto()
    # ASPIRATING = auto()   # PICKED_UP_TIP으로 간주



import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.common.suitable.enums import Axis

PIPETTE_COUNT: int = 2
DECK_COUNT: int = 12

PIPETTE_NUMBERS = [i + 1 for i in range(PIPETTE_COUNT)]
DECK_NUMBERS = [i + 1 for i in range(DECK_COUNT)]

AXIS_ZS = [Axis.Z[n] for n in PIPETTE_NUMBERS]
AXIS_PS = [Axis.P[n] for n in PIPETTE_NUMBERS]



import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.network.messenger import MessengerClient, run_server_func
from ablelabs.neon.utils.network.tcp_client import TcpClient
from ablelabs.neon.controllers.notable.nanophilia.api.robot_router import RobotRouter


class SetAPI(MessengerClient):
    def __init__(self, tcp_client: TcpClient) -> None:
        super().__init__(tcp_client)

    @run_server_func(RobotRouter.set_pipettes)
    async def pipettes(self, value: dict[int, str | None]):
        pass

    @run_server_func(RobotRouter.set_tips)
    async def tips(self, value: dict[int, str | None]):
        pass

    @run_server_func(RobotRouter.set_deck_modules)
    async def deck_modules(self, value: dict[int, str | None]):
        pass

    @run_server_func(RobotRouter.set_labwares)
    async def labwares(self, value: dict[int, str | None]):
        pass

    @run_server_func(RobotRouter.set_update_setup_data)
    async def update_setup_data(self, value: dict, save: bool):
        pass

    @run_server_func(RobotRouter.set_update_pipette_attrs)
    async def update_pipette_attrs(self, value: dict[int, dict]):
        pass



from loguru import logger
import asyncio
import time

import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.network.base_tcp import *


class TcpServer(BaseTcp):
    def __init__(
        self,
        name="Server",
        encoding: str = "utf-8",
        log_func: Callable = logger.trace,
    ):
        super().__init__(name=name, encoding=encoding, log_func=log_func)
        self._server: asyncio.Server = None
        self.on_disconnected = None

    async def open(self, ip, port):
        logger.info(f"Waiting for a connection...({ip}:{port})")
        try:
            self._server = await asyncio.start_server(
                self._client_connected_cb, ip, port
            )
        except Exception as e:
            self._is_connected = False

    async def _client_connected_cb(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ):
        self._reader = reader
        self._writer = writer
        client_address = writer.get_extra_info("peername")
        logger.info(f"Connection from {client_address}")
        asyncio.create_task(self.handle_receive())
        self._is_connected = True
        for on_connected in self.on_connected:
            await on_connected()

    async def close(self):
        if not self._server:
            return
        self._server.close()
        await self._server.wait_closed()


async def main():
    # TcpServer.example()

    logger.remove()
    logger.add(sys.stdout, level="TRACE")
    # logger.add(sys.stdout, level="DEBUG")
    logger.add("logs/trace.log", level="TRACE")
    logger.add("logs/debug.log", level="DEBUG")
    logger.add("logs/info.log", level="INFO")

    ip = "localhost"
    port = 1234

    tcp_server = TcpServer(name="tcp_server", log_func=logger.trace)

    async def on_server_received(message: str):
        logger.debug(f"server <- {message}")

    tcp_server.on_received.append(on_server_received)
    await tcp_server.open(ip=ip, port=port)

    from ablelabs.neon.utils.network.tcp_client import TcpClient

    tcp_client = TcpClient(name="tcp_client", log_func=logger.trace)

    async def on_client_received(message: str):
        logger.debug(f"client <- {message}")

    tcp_client.on_received.append(on_client_received)
    await tcp_client.connect(ip=ip, port=port)

    await tcp_server.send("server send to client")
    await tcp_client.send("client send to server")


if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    loop.create_task(main())
    loop.run_forever()



import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.network.messenger import MessengerClient, run_server_func
from ablelabs.neon.utils.network.tcp_client import TcpClient
from ablelabs.neon.controllers.alma.api.robot_router import RobotRouter
from ablelabs.neon.controllers.alma.api.time.upper_module import TimeUpperModuleAPI


class TimeAPI(MessengerClient):
    def __init__(self, tcp_client: TcpClient) -> None:
        super().__init__(tcp_client)
        self.upper_module = TimeUpperModuleAPI(tcp_client)

    @run_server_func(RobotRouter.time_initialize)
    async def initialize(
        self,
        # recovery: bool,
    ) -> float:
        pass

    @run_server_func(RobotRouter.time_calibrate_lcr_meter)
    async def calibrate_lcr_meter(self) -> float:
        pass



from loguru import logger

import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.network.messenger import MessengerClient, run_server_func
from ablelabs.neon.utils.network.tcp_client import TcpClient
from ablelabs.neon.controllers.notable.kribb.api.robot_router import RobotRouter
from ablelabs.neon.controllers.notable.api.motion_api import MotionAPI as BaseMotionAPI
from ablelabs.neon.common.notable.structs import Location


class MotionAPI(BaseMotionAPI):
    def __init__(self, tcp_client: TcpClient) -> None:
        super().__init__(tcp_client)

    @run_server_func(RobotRouter.motion_move_to_camera)
    async def move_to_camera(
        self,
        pipette_number: int,
        location: Location,
    ):
        pass

    @run_server_func(RobotRouter.motion_move_to_displacement_sensor)
    async def move_to_displacement_sensor(
        self,
        pipette_number: int,
        location: Location,
    ):
        pass



import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.enum_types import StrEnum, auto, IntFlag, IntEnum, ItemGetable


class NUMBERING_ORDER(StrEnum):
    LTR_TTB = auto()  # Left to Right, Top to Bottom
    LTR_BTT = auto()  # Left to Right, Bottom to Top
    TTB_LTR = auto()  # Top to Bottom, Left to Right
    BTT_LTR = auto()  # Bottom to Top, Left to Right


class Color(IntFlag):
    NONE = 0
    RED = 1 << 0
    GREEN = 1 << 1
    BLUE = 1 << 2
    WHITE = 1 << 3
    CYAN = GREEN | BLUE
    MAGENTA = RED | BLUE
    YELLOW = RED | GREEN


class RunStatus(IntEnum):
    BOOT = auto()
    SHUTDOWN = auto()
    READY = auto()
    DONE = auto()
    INITIALIZE = auto()
    RUN = auto()
    PAUSE = auto()
    # 의도한 상황
    STOP = auto()
    EMERGENCY = auto()
    # 예기치 못한 상황
    ERROR = auto()  # 일반적인 에러
    TIMEOUT = auto()
    FAULT = auto()


class TaskStatus(StrEnum):
    RUNNING = auto()
    PAUSED = auto()
    STOPPED = auto()


class Axis(StrEnum):
    X = auto()
    Y = auto()
    Z1 = auto()
    Z2 = auto()
    P1 = auto()
    P2 = auto()

    @staticmethod
    def Z_(pipette_number: int):
        return Axis.from_str(f"{Axis.Z1[0]}{pipette_number}")

    @staticmethod
    def P_(pipette_number: int):
        return Axis.from_str(f"{Axis.P1[0]}{pipette_number}")

    Z = ItemGetable[int](lambda number: Axis.from_str(f"{Axis.Z1[0]}{number}"))
    P = ItemGetable[int](lambda number: Axis.from_str(f"{Axis.P1[0]}{number}"))


class LabwareType(StrEnum):
    TIP_RACK = auto()
    WASTE = auto()
    WELL_PLATE = auto()

    def __eq__(self, __value: object) -> bool:
        return super().__eq__(str(__value).upper())


class Height(StrEnum):
    UP = auto()
    TOP = auto()
    TOP_JUST = auto()
    BOTTOM = auto()
    BOTTOM_JUST = auto()
    TIP_RACK_TOUCHED = auto()
    TIP_RACK_PRESSED = auto()
    WASTE_DROP_TIP = auto()
    WASTE_BLOW_OUT = auto()


class Plunger(StrEnum):
    HOME_OFFSET = auto()
    DROP_TIP = auto()
    TOP = auto()
    FIRST_STOP = auto()
    SECOND_STOP = auto()


class LocationType(StrEnum):
    DECK = auto()


class LocationReference(StrEnum):
    TOP = auto()
    TOP_JUST = auto()
    BOTTOM = auto()
    BOTTOM_JUST = auto()



from setuptools import setup, find_packages

setup(
    name="ablelabs",
    version="0.3.6",
    description="",
    author="sypark",
    author_email="sy.park@ablelabsinc.com",
    url="https://github.com/ABLE-Labs/ABLE-API",
    install_requires=[
        "et-xmlfile>=1.1.0",
        "future>=1.0.0",
        "iso8601>=2.1.0",
        "loguru>=0.7.2",
        "openpyxl>=3.1.5",
        "pyserial>=3.5",
        "PyYAML>=6.0.1",
    ],
    packages=find_packages(exclude=[]),
    keywords=["ablelabs"],
    python_requires=">=3.10",
    package_data={},
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3.10",
    ],
)



from dataclasses import dataclass
import time
from typing import Callable
import asyncio
import inspect
from functools import update_wrapper
import numpy as np


def log_elapsed_time(log_func: Callable):
    def decorator(func: Callable):
        def wrapper(*args, **kwargs):
            start_time = time.perf_counter()
            result = func(*args, **kwargs)
            elapsed_time = time.perf_counter() - start_time
            log_func(f"{func.__name__} : {elapsed_time*1e3:.1f} ms")
            return result

        update_wrapper(wrapper, func)
        return wrapper

    return decorator


def log_elapsed_time_async(log_func: Callable):
    def decorator(func: Callable):
        async def wrapper(*args, **kwargs):
            start_time = time.perf_counter()
            result = await func(*args, **kwargs)
            elapsed_time = time.perf_counter() - start_time
            log_func(f"{func.__name__} : {elapsed_time*1e3:.1f} ms")
            return result

        update_wrapper(wrapper, func)
        return wrapper

    return decorator


@log_elapsed_time(print)
def test_check_time(n):
    for _ in range(n):
        time.sleep(0.01)


@log_elapsed_time_async(print)
async def test_check_time_async(n):
    for _ in range(n):
        await asyncio.sleep(0.01)


def get_args_dict(func: Callable, *args, **kwargs):
    sig = inspect.signature(func)
    try:  # sypark 여기 try문을 다른 걸로 대체할 수 없을까?
        bound = sig.bind(*args, **kwargs)
    except:
        bound = sig.bind(None, *args, **kwargs)
    # if "self" in sig.parameters:
    #     bound = sig.bind(None, *args, **kwargs)
    # else:
    #     bound = sig.bind(*args, **kwargs)
    bound.apply_defaults()
    result = dict(bound.arguments)
    if "self" in result and result["self"] is None:
        del result["self"]
    return result


def str_value(value, max_len: int = None, level: int = 0, max_level: int = 0):
    """
    객체를 문자열로 변환하며, 너무 긴 경우 줄임표로 요약하는 함수.

    Args:
    - value: 변환할 값 (list, dict, tuple, 기본 타입 등)
    - max_len: 출력할 최대 길이
    - level: 현재 중첩 수준 (기본값: 0)
    - max_level: 문자열을 줄이기 시작할 중첩 수준 (기본값: 1)

    Returns:
    - str: 요약된 문자열 표현
    """

    # 중첩 수준에 따른 문자열 요약 처리
    if isinstance(value, list):
        # 리스트일 경우
        value_str = (
            "["
            + ", ".join(str_value(v, max_len, level + 1, max_level) for v in value)
            + "]"
        )
    elif isinstance(value, dict):
        # 딕셔너리일 경우
        value_str = (
            "{"
            + ", ".join(
                f"{str_value(k, max_len, level + 1, max_level)}: {str_value(v, max_len, level + 1, max_level)}"
                for k, v in value.items()
            )
            + "}"
        )
    elif isinstance(value, tuple):
        # 튜플일 경우
        value_str = (
            "("
            + ", ".join(str_value(v, max_len, level + 1, max_level) for v in value)
            + ")"
        )
    elif inspect.iscoroutine(value):
        # Coroutine 객체의 경우
        value_str = value.__qualname__
    elif isinstance(value, asyncio.Task):
        # 비동기 Task 객체의 경우
        value_str = value.get_coro().__qualname__
    elif isinstance(value, np.ndarray):
        height, width = value.shape[:2]
        channels = value.shape[2] if len(value.shape) == 3 else 1
        data_type = value.dtype
        value_str = f"{height}x{width}({channels}ch,{data_type})"
    else:
        # 기본 타입일 경우
        value_str = str(value)

    # 줄바꿈 문자를 제거하고 공백으로 대체
    value_str = value_str.replace("\r", "").replace("\n", "")

    # 중첩 수준에 따라 문자열을 자르고, 요약된 길이를 출력
    if max_len and len(value_str) > max_len and level >= max_level:
        value_str = value_str[:max_len] + " ..."
        if isinstance(value, list):
            value_str += f" (len: {len(value)})"
        elif isinstance(value, dict):
            value_str += f" (len: {len(value)})"
        elif isinstance(value, tuple):
            value_str += f" (len: {len(value)})"

    return value_str


def get_dict_str(value: dict):
    args_strs = []
    for key, value in value.items():
        args_strs.append(f"{key}={str_value(value)}")
    result = f"{' '.join(args_strs)}"
    return result


def get_func_args_str(func: Callable, args_dict: dict, max_len: int = None):
    # args_strs = [f"{k}={v}" for k, v in args_dict.items() if k != "self"]
    args_strs = []
    for key, value in args_dict.items():
        if key == "self":
            continue
        args_strs.append(f"{key}={str_value(value, max_len=max_len)}")
    result = f"{func.__name__}({' '.join(args_strs)})"
    return result


def log_func_args(
    log_before_func: Callable = None,
    log_after_func: Callable = None,
    max_len: int = None,
):
    def decorator(func: Callable):
        def wrapper(*args, **kwargs):
            args_dict = get_args_dict(func, *args, **kwargs)
            func_args_strs = get_func_args_str(func, args_dict, max_len)
            if log_before_func:
                log_before_func(f"{func_args_strs}")
            try:
                result = func(*args, **kwargs)
            except Exception as e:
                if log_before_func:
                    log_before_func(f"{func_args_strs} : {e}")
                elif log_after_func:
                    log_after_func(f"{func_args_strs} : {e}")
                raise e
            if log_after_func:
                log_after_func(f"{func_args_strs} -> {str_value(result, max_len=50)}")
            return result

        update_wrapper(wrapper, func)
        return wrapper

    return decorator


def log_func_args_async(
    log_before_func: Callable = None,
    log_after_func: Callable = None,
    max_len: int = None,
):
    def decorator(func: Callable):
        async def wrapper(*args, **kwargs):
            args_dict = get_args_dict(func, *args, **kwargs)
            func_args_strs = get_func_args_str(func, args_dict, max_len)
            if log_before_func:
                log_before_func(f"{func_args_strs}")
            try:
                result = await func(*args, **kwargs)
            except Exception as e:
                if log_before_func:
                    log_before_func(f"{func_args_strs} : {e}")
                elif log_after_func:
                    log_after_func(f"{func_args_strs} : {e}")
                raise e
            if log_after_func:
                log_after_func(f"{func_args_strs} -> {str_value(result, max_len=50)}")
            return result

        update_wrapper(wrapper, func)
        return wrapper

    return decorator


# @log_elapsed_time_async(print)
@log_func_args_async(log_before_func=print, log_after_func=print)
async def test_log_func_args_async(a: int, b: float, c: str):
    await asyncio.sleep(0.01)


def skip_func_async(skip_func: Callable[[], bool], return_value=None):
    def decorator(func):
        async def wrapper(*args, **kwargs):
            if skip_func():
                return return_value
            return await func(*args, **kwargs)

        update_wrapper(wrapper, func)
        return wrapper

    return decorator


@skip_func_async(skip_func=lambda: True, return_value="return_sample")
async def test_skip_func_async_do_skip():
    print(f"{test_skip_func_async_do_skip.__name__}")


@skip_func_async(skip_func=lambda: False, return_value="return_sample")
async def test_skip_func_async_donot_skip():
    print(f"{test_skip_func_async_donot_skip.__name__}")


@log_func_args(log_before_func=print, log_after_func=print)
def test_long_args(v):
    pass


async def main():
    # test_check_time(1)
    # test_check_time(2)
    # test_check_time(3)

    # await test_check_time_async(1)
    # await test_check_time_async(2)
    # await test_check_time_async(3)

    # await test_log_func_args_async(a=1, b=2.3, c="abc")
    # await test_log_func_args_async(a=0, b=0, c="")
    # await test_log_func_args_async(1, 2.3, "abc")

    # print(await test_skip_func_async_do_skip())
    # print(await test_skip_func_async_donot_skip())

    # num_contours = 5
    # max_points = 100
    # img_size = (500, 500)
    # contours = []
    # for _ in range(num_contours):
    #     # 윤곽선을 구성할 점의 수를 임의로 결정 (최소 4개 이상의 점을 가진 다각형 생성)
    #     num_points = np.random.randint(4, max_points)
    #     # 윤곽선을 구성할 (x, y) 좌표를 임의로 생성
    #     points = np.random.randint(0, min(img_size), size=(num_points, 1, 2))
    #     # 생성된 좌표를 numpy 배열로 변환하여 contours에 추가
    #     contours.append(points.astype(np.int32))
    # test_long_args(contours)

    @dataclass
    class Info:
        a: int
        b: tuple[float, float]
        c: list[str]

    info = Info(a=1, b=(2.3, 3.4), c=[f"{v}" for v in range(1000)])
    infos = [info for _ in range(10)]
    test_long_args(info)
    test_long_args(infos)


if __name__ == "__main__":
    asyncio.run(main())
